from stream import Stream
from counter import Counter
from helper import Helper
from sink import Sink

class ComponentStream(Stream):
  def sink(self, name=None, fields=None):
    self.outputs(name, fields)

  def outputs(self, name=None, fields=None, relation=None, scope=None):
    node = Sink(name, fields, relation, scope)
    Helper.check_name("outputs", node._name, self._app._names)
    Helper.check_component_sink(node)
    self._app._nodes.append(node)

    if(self._app._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type, "columns":node._columns, "relation":node._relation, "scope":node._scope}
      Helper.write_node_to_file(node_hash, self._app._info_file)
      arc_hash = {"name":self._name, "origin":self._previous_node_name, "dest":node._name}
      Helper.write_arc_to_file(arc_hash, self._app._info_file)



import traceback
from counter import Counter
from controller import ParentDeadException

class Filter:
  def __init__(self, name, emits, prepare, keep, parallelism):
    self._name = name if name else "filter_"+Counter.get()
    self._type = "filter"

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._prepare = prepare
    self._keep = keep
    self._parallelism = parallelism


  def run(self, controller):
    controller.get_pidDir()
    if self._prepare != None:
      self._prepare(controller)

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        tup = controller.get_tuple(d)
        if tup == None:
          continue
        if self._keep(tup):
          controller.emit(tup)
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in filter: "+traceback.format_exc())
      controller.done()



import traceback
from counter import Counter
from controller import ParentDeadException, UserEndCycleException

class Source:
  def __init__(self, name, matches, options, emits, end_cycle_policy, begin_cycle, next_tuple):
    self._name = name if name else "source_"+Counter.get()
    self._type = "source"
    self._matches = None
    self._relation = None
    self._end_cycle_policy = None

    if(matches != None):
      if isinstance(matches, basestring):
        if not isinstance(options, dict):
          options = {}
        self._relation = {"query": matches, "options": options}
      elif isinstance(matches, list):
        self._matches = matches
    else:
      self._end_cycle_policy = end_cycle_policy

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._begin_cycle = begin_cycle
    self._next_tuple = next_tuple

  def run(self, controller):
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" not in d:
          controller.log("Not a command")
        elif d["command"] == "begin_cycle":
          if(self._begin_cycle != None):
            self._begin_cycle(controller)
        elif d["command"] == "next":
            self._next_tuple(controller)
        elif d["command"] == "ack":
          controller.ack(d.get("id",None))
        elif d["command"] == "fail":
          controller.fail(d.get("id",None))
        else:
          controller.log("Invalid command")
      except UserEndCycleException, e:
        pass
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in source: "+traceback.format_exc())
      controller.done()



from counter import Counter

class InjectedComponent:
  def __init__(self, name, idn, additional_inputs, outputs, output_format):
    self._name = name if name else "component_"+Counter.get()
    self._id = str(idn)
    self._type = "component"
    self._consumes = []

    if isinstance(additional_inputs, list):
      self._consumes = additional_inputs
    elif additional_inputs != None:
      self._consumes = [additional_inputs]

    if outputs == None:
      outputs = ["stream_"+Counter.get()]
    elif isinstance(outputs, basestring):
      outputs = [outputs]
    self._emits = outputs

    self._output_format = output_format



from counter import Counter

class Join:
  def __init__(self, name, emits, fields, options):
    self._name = name if name else "join_"+Counter.get()
    self._type = "join"

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    if fields:
      if isinstance(fields, list) and len(fields) == 2:
        self._lhs_fields = fields[0]
        self._rhs_fields = fields[1]
      elif isinstance(fields, basestring):
        self._lhs_fields = fields
        self._rhs_fields = fields

    self._join_type = "inner"
    if options:
      if options["type"]:
        self._join_type = options["type"]



from controller import *
from counter import Counter
from helper import Helper
from stream import *
from source import Source
from base import Base
from version import VERSION

class App(Base):
  def __init__(self, name=None):
    Base.__init__(self, name)
    Helper.check_name("app", self._name, {})
    self._options = Helper.opt_parser()

    if(self._options["command"] == "info"):
      self._info_file = open(self._options["file"], "w+")
      Helper.write_hash_to_file({"language":"python", "name":self._name, "flow_type":"app", "multilang_version":VERSION}, self._info_file)

  def source(self, name=None, matches=None, options={}, emits=None, end_cycle_policy="null_emit", begin_cycle=None, next_tuple=None):
    if(matches != None):
      Helper.check_matches(matches)
    node = Source(name, matches, options, emits, end_cycle_policy, begin_cycle, next_tuple)
    Helper.check_name("source", node._name, self._names)
    Helper.check_source(node)
    Helper.check_emits("source", node._emits, self._streams)
    self._nodes.append(node)

    if(self._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type}
      if(node._relation):
        node_hash["relation"] = node._relation
      elif(node._matches):
        node_hash["matches"] = node._matches
      else:
        node_hash["end_cycle_policy"] = node._end_cycle_policy
      Helper.write_node_to_file(node_hash, self._info_file)
    elif self._options["command"] == "execute" and self._options["name"] == node._name:
      pipe_name = self._options.get("pipe","")
      controller = Controller(pipe_name, node._emits)
      node.run(controller)

    output_streams = []
    for stream in node._emits:
      output_streams.append(Stream(stream, self, node._name))
    if len(output_streams) == 1:
      output_streams = output_streams[0]

    return output_streams



import traceback
from counter import Counter
from controller import ParentDeadException

class Each:

  def __init__(self, name, emits, output_format, prepare, execute, parallelism):
    self._name = name if name else "each_"+Counter.get()
    self._type = "each"
    self._output_format = output_format

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._prepare = prepare
    self._execute = execute
    self._parallelism = parallelism

  def run(self, controller):
    controller.get_pidDir()
    if self._prepare != None:
      self._prepare(controller)

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        tup = controller.get_tuple(d)
        if tup == None:
          continue
        self._execute(controller, tup)
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in each: "+traceback.format_exc())
      controller.done()



from counter import Counter

class ComponentSource:
  def __init__(self, name, fields):
    self._name = name
    self._type = "source"

    self._emits = ["stream_"+Counter.get()]
    self._fields = fields



import traceback
from controller import *
from counter import Counter
from helper import Helper
from each import Each
from filter import Filter
from group_by import GroupBy
from join import Join
from injected_component import InjectedComponent
from sink import Sink

class Stream:
  def __init__(self, name, app, previous_node_name):
    self._name = name
    self._app = app
    self._previous_node_name = previous_node_name

  def each(self, name=None, emits=None, output_format="replace", prepare=None, execute=None, parallelism=None):
    node = Each(name, emits, output_format, prepare, execute, parallelism)

    Helper.check_name("each", node._name, self._app._names)
    Helper.check_each(node)
    Helper.check_emits("each", node._emits, self._app._streams)
    self._app._nodes.append(node)

    if(self._app._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type, "output_format":node._output_format}

      if node._parallelism != None:
        node_hash["parallelism"] = node._parallelism

      Helper.write_node_to_file(node_hash, self._app._info_file)
      arc_hash = {"name":self._name, "origin":self._previous_node_name, "dest":node._name}
      Helper.write_arc_to_file(arc_hash, self._app._info_file)
    elif self._app._options["command"] == "execute" and self._app._options["name"] == node._name:
      pipe_name = self._app._options.get("pipe","")
      controller = Controller(pipe_name, node._emits)
      node.run(controller)

    output_streams = []
    for stream in node._emits:
      output_streams.append(self.__class__(stream, self._app, node._name))
    if len(output_streams) == 1:
      output_streams = output_streams[0]

    return output_streams

  def filter(self, name=None, emits=None, prepare=None, keep=None, parallelism=None):
    node = Filter(name, emits, prepare, keep, parallelism)
    Helper.check_name("filter", node._name, self._app._names)
    Helper.check_filter(node)
    Helper.check_emits("filter", node._emits, self._app._streams)
    self._app._nodes.append(node)

    if(self._app._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type}
      
      if node._parallelism != None:
        node_hash["parallelism"] = node._parallelism

      Helper.write_node_to_file(node_hash, self._app._info_file)
      arc_hash = {"name":self._name, "origin":self._previous_node_name, "dest":node._name}
      Helper.write_arc_to_file(arc_hash, self._app._info_file)
    elif self._app._options["command"] == "execute" and self._app._options["name"] == node._name:
      pipe_name = self._app._options.get("pipe", "")
      controller = Controller(pipe_name, node._emits)
      node.run(controller)

    output_stream = self.__class__(node._emits[0], self._app, node._name) # filters only have one output stream
    return output_stream

  def group_by(self, name=None, emits=None, fields=None, begin_group=None, aggregate=None, end_group=None, parallelism=None):
    node = GroupBy(name, emits, fields, begin_group, aggregate, end_group, parallelism)
    Helper.check_name("group_by", node._name, self._app._names)
    Helper.check_group_by(node)
    Helper.check_emits("group_by", node._emits, self._app._streams)
    self._app._nodes.append(node)

    if(self._app._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type, "group_by":node._group_by}
      if node._parallelism != None:
        node_hash["parallelism"] = node._parallelism
        
      Helper.write_node_to_file(node_hash, self._app._info_file)
      arc_hash = {"name":self._name, "origin":self._previous_node_name, "dest":node._name}
      Helper.write_arc_to_file(arc_hash, self._app._info_file)
    elif self._app._options["command"] == "execute" and self._app._options["name"] == node._name:
      pipe_name = self._app._options.get("pipe","")
      controller = Controller(pipe_name, node._emits)
      node.run(controller)

    output_streams = []
    for stream in node._emits:
      output_streams.append(self.__class__(stream, self._app, node._name))
    if len(output_streams) == 1:
      output_streams = output_streams[0]

    return output_streams

  def join_with(self, name=None, stream=None, emits=None, fields=None, options=None):
    node = Join(name, emits, fields, options)
    Helper.check_name("join", node._name, self._app._names)
    Helper.check_join(node)
    Helper.check_emits("join", node._emits, self._app._streams)
    self._app._nodes.append(node)

    if(self._app._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type, "lhs_fields":node._lhs_fields, "rhs_fields":node._rhs_fields, "join_type":node._join_type}
         
      Helper.write_node_to_file(node_hash, self._app._info_file)
      arc_hash1 = {"name":self._name, "origin":self._previous_node_name, "dest":node._name, "left":1}
      Helper.write_arc_to_file(arc_hash1, self._app._info_file)
      arc_hash2 = {"name":stream._name, "origin":stream._previous_node_name, "dest":node._name, "right":1}
      Helper.write_arc_to_file(arc_hash2, self._app._info_file)

    output_stream = self.__class__(node._emits[0], self._app, node._name)
    return output_stream

  def call_component(self, name=None, component_id=None, additional_inputs=None, outputs=None, output_format="replace"):
    node = InjectedComponent(name, component_id, additional_inputs, outputs, output_format)
    Helper.check_name("component", node._name, self._app._names)
    Helper.check_call_component(node)
    Helper.check_emits("component", node._emits, self._app._streams)
    self._app._nodes.append(node)

    if(self._app._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type, "id":node._id, "output_format":node._output_format}
      Helper.write_node_to_file(node_hash, self._app._info_file)
      arc_hash = {"name":self._name, "origin":self._previous_node_name, "dest":node._name}
      Helper.write_arc_to_file(arc_hash, self._app._info_file)
      for stream in node._consumes:
        arc_hash = {"name":stream._name, "origin":stream._previous_node_name, "dest":node._name}
        Helper.write_arc_to_file(arc_hash, self._app._info_file)

    output_streams = []
    for stream in node._emits:
      output_streams.append(self.__class__(stream, self._app, node._name))
    if len(output_streams) == 1:
      output_streams = output_streams[0]

    return output_streams 

  def sink(self, name=None, columns=None, relation=None, scope=None):
    node = Sink(name, columns, relation, scope)
    Helper.check_sink(node._name, node._columns, self._app._nodes)
    self._app._nodes.append(node)

    if(self._app._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type, "columns":node._columns, "relation":node._relation, "scope":node._scope}
      Helper.write_node_to_file(node_hash, self._app._info_file)
      arc_hash = {"name":self._name, "origin":self._previous_node_name, "dest":node._name}
      Helper.write_arc_to_file(arc_hash, self._app._info_file)



class Sink:
  def __init__(self, name, columns, relation, scope):
    self._name = name
    self._type = "sink"
    self._columns = columns

    if relation != None:
      self._relation = relation
    else:
      self._relation = name
    self._scope = scope



from controller import *
from counter import Counter
from helper import Helper
from component_stream import ComponentStream
from component_source import ComponentSource
from base import Base
from version import VERSION

class Component(Base):
  def __init__(self, name=None):
    Base.__init__(self, name)
    Helper.check_name("component", self._name, {})
    self._options = Helper.opt_parser()

    if(self._options["command"] == "info"):
      self._info_file = open(self._options["file"], "w+")
      Helper.write_hash_to_file({"language":"python", "name":self._name, "flow_type":"component", "multilang_version":VERSION}, self._info_file)

  def inputs(self, name=None, fields=None):
    node = ComponentSource(name, fields)
    Helper.check_name("inputs", node._name, self._names)
    Helper.check_component_source(node)

    self._nodes.append(node)

    if(self._options["command"] == "info"):
      node_hash = {"name":node._name, "type":node._type, "fields":node._fields}
      Helper.write_node_to_file(node_hash, self._info_file)

    # component inputs only specify one output stream at a time
    output_stream = ComponentStream(node._emits[0], self, node._name)
    return output_stream



import traceback
from counter import Counter
from controller import ParentDeadException

class GroupBy:
  def __init__(self, name, emits, fields, begin_group, aggregate, end_group, parallelism):
    self._name = name if name else "group_by_"+Counter.get()
    self._type = "group_by"

    self._group_by = fields
    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._begin_group = begin_group
    self._aggregate = aggregate
    self._end_group = end_group
    self._parallelism = parallelism

  def run(self, controller):
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" not in d:
          controller.log("Not a command "+str(d))
        elif d["command"] == "begin_group":
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._begin_group(controller, tup)
        elif d["command"] == "aggregate":
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._aggregate(controller, tup)
        elif d["command"] == "end_group":
          self._end_group(controller)
        else:
          controller.log("Invalid command")
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except KeyboardInterrupt:
        raise
      except:
        controller.fail("Exception in group_by: "+traceback.format_exc())
      controller.done()
